/* 
 * This utility converts Macintosh's line termination characters into a UNIX
 * <LF> line termination style. The result is written to standard output 
 * (stdout) and may be redirected to a file.
 *
 * Written by Pavol Dilung -- <neuron.tuke.sk> 2003
 *
 * USAGE: mac2unix <filename> [ > <desired_output> ]
 */

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

int main(int argc, char *argv[])
{
    FILE *fr;

    if (argc != 2) {
        fprintf(stderr, "Usage: %s <filename>\n", argv[0]);
        exit(EXIT_FAILURE);
    }
    
    if (! (fr = fopen(argv[1], "r"))) {
        fprintf(stderr, "Error: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }
    
    while (! feof(fr)) {
        char c;
        
        if (! (c = fgetc(fr))) {
            fprintf(stderr, "Error: %s\n", strerror(errno));
            exit(EXIT_FAILURE);
        }

        if (c == 0x0D || c == '\377') c = '\n';

        putchar(c);
    }

    if (fclose(fr) == EOF) {
        fprintf(stderr, "Error: %s\n", strerror(errno));
        exit(EXIT_FAILURE);
    }

    return (0);
}

